@@ -1,5 +1,5 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
-/**
+/*
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -33,7 +33,6 @@
 #include <gtk/gtk.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-mime-utils.h>
-#include <libgnomevfs/gnome-vfs-uri.h>
 #include <gconf/gconf-client.h>
 #include "egg-recent-model.h"
 #include "egg-recent-item.h"
@@ -390,21 +389,23 @@ end_element_handler (GMarkupParseContext *context,
 			    GError **error)
 {
 	ParseInfo *info = (ParseInfo *)user_data;
-	GnomeVFSURI *uri;
 
 	switch (peek_state (info)) {
 		case STATE_RECENT_ITEM:
-			if (info->current_item->uri == NULL ||
-			    strlen (info->current_item->uri) == 0) {
-				egg_recent_item_unref(info->current_item);
-				info->current_item = NULL;
-				break;
-			}
+                        if ((info->current_item->mime_type == NULL) &&
+                            !info->current_item->mime_type_is_explicit)
+                        {
+                                egg_recent_item_set_mime_type (info->current_item, NULL);
+                        }
+ 
 			info->items = g_list_append (info->items,
-						     info->current_item);
-			break;
+						    info->current_item);
+			if (info->current_item->uri == NULL ||
+			    strlen (info->current_item->uri) == 0)
+				g_warning ("URI NOT LOADED");
+		break;
 		default:
-			break;
+		break;
 	}
 
 	pop_state (info);
@@ -427,7 +428,17 @@ text_handler (GMarkupParseContext *context,
 		case STATE_GROUPS:
 		break;
 		case STATE_URI:
-			egg_recent_item_set_uri (info->current_item, text);
+                {
+                        if ((info->current_item->mime_type == NULL) &&
+                            !info->current_item->mime_type_is_explicit)
+                        {
+                                info->current_item->mime_type_is_explicit = TRUE;
+                                egg_recent_item_set_uri (info->current_item, text);
+                                info->current_item->mime_type_is_explicit = FALSE;
+                        }
+                        else
+                                egg_recent_item_set_uri (info->current_item, text);
+                }
 		break;
 		case STATE_MIME_TYPE:
 			egg_recent_item_set_mime_type (info->current_item,
@@ -523,7 +534,6 @@ egg_recent_model_filter (EggRecentModel *model,
 	GList *newlist = NULL;
 	gchar *mime_type;
 	gchar *uri;
-	GnomeVFSURI *vfs_uri;
 
 	g_return_val_if_fail (list != NULL, NULL);
 
@@ -531,7 +541,6 @@ egg_recent_model_filter (EggRecentModel *model,
 		gboolean pass_mime_test = FALSE;
 		gboolean pass_group_test = FALSE;
 		gboolean pass_scheme_test = FALSE;
-
 		item = (EggRecentItem *)list->data;
 		list = list->next;
 
@@ -575,16 +584,8 @@ egg_recent_model_filter (EggRecentModel *model,
 		} else
 			pass_scheme_test = TRUE;
 
-
-		if (pass_mime_test && pass_group_test && pass_scheme_test) {
-			/* filter out non-existent URI-s */
-			vfs_uri = gnome_vfs_uri_new(uri);
-			if(vfs_uri) {
-				if(gnome_vfs_uri_exists(vfs_uri))
-					newlist = g_list_prepend
-						(newlist, item);
-			}
-		}
+		if (pass_mime_test && pass_group_test && pass_scheme_test)
+			newlist = g_list_prepend (newlist, item);
 
 		g_free (uri);
 	}
@@ -594,6 +595,7 @@ egg_recent_model_filter (EggRecentModel *model,
 		g_list_free (list);
 	}
 
+	
 	return newlist;
 }
 
@@ -691,12 +693,17 @@ static void
 egg_recent_model_monitor (EggRecentModel *model, gboolean should_monitor)
 {
 	if (should_monitor && model->priv->monitor == NULL) {
+		char *uri;
+
+		uri = gnome_vfs_get_uri_from_local_path (model->priv->path);
 
 		gnome_vfs_monitor_add (&model->priv->monitor,
-					     model->priv->path,
-					     GNOME_VFS_MONITOR_FILE,
-					     egg_recent_model_monitor_cb,
-					     model);
+				       uri,
+				       GNOME_VFS_MONITOR_FILE,
+				       egg_recent_model_monitor_cb,
+				       model);
+
+		g_free (uri);
 
 		/* if the above fails, don't worry about it.
 		 * local notifications will still happen
@@ -871,14 +878,17 @@ static FILE *
 egg_recent_model_open_file (EggRecentModel *model)
 {
 	FILE *file;
+	mode_t prev_umask;
 	
 	file = fopen (model->priv->path, "r+");
 	if (file == NULL) {
 		/* be paranoid */
-		umask (077);
+		prev_umask = umask (077);
 
 		file = fopen (model->priv->path, "w+");
 
+		umask (prev_umask);
+
 		g_return_val_if_fail (file != NULL, NULL);
 	}
 
@@ -889,11 +899,34 @@ static gboolean
 egg_recent_model_lock_file (FILE *file)
 {
 	int fd;
+	gint	try = 5;
 
 	rewind (file);
 	fd = fileno (file);
 
-	return lockf (fd, F_LOCK, 0) == 0 ? TRUE : FALSE;
+	/* Attempt to lock the file 5 times,
+	 * waiting a random interval (< 1 second) 
+	 * in between attempts.
+	 * We should really be doing asynchronous
+	 * locking, but requires substantially larger
+	 * changes.
+	 */
+	
+	while (try > 0)
+	{
+		int rand_interval;
+
+		if (lockf (fd, F_TLOCK, 0) == 0)
+			return TRUE;
+
+		rand_interval = 1 + (int) (10.0 * rand()/(RAND_MAX + 1.0));
+			 
+           	g_usleep (100000 * rand_interval);
+
+		--try;
+	}
+
+	return FALSE;
 }
 
 static gboolean
@@ -904,7 +937,7 @@ egg_recent_model_unlock_file (FILE *file)
 	rewind (file);
 	fd = fileno (file);
 
-	return lockf (fd, F_ULOCK, 0) < 0 ? FALSE : TRUE;
+	return (lockf (fd, F_ULOCK, 0) == 0) ? TRUE : FALSE;
 }
 
 static void
@@ -1274,6 +1307,7 @@ egg_recent_model_add_full (EggRecentModel * model, EggRecentItem *item)
 		ret = TRUE;
 	} else {
 		g_warning ("Failed to lock:  %s", strerror (errno));
+		fclose (file);
 		return FALSE;
 	}
 
